from typing import Annotated, Any
from pydantic.fields import PydanticUndefined 
from pydantic.functional_validators import AfterValidator
from pydantic import Field

from .. import ConfigBase


def MutuallyExclusive(typ: type, default: Any = PydanticUndefined, ):
    def 

    return Annotated[typ, AfterValidator(...)]


if __name__ == "__main__":
    class Config(ConfigBase):
        a: bool = 



import inspect
from typing import Annotated, Generic, TypeVar
from pydantic import Field


T = TypeVar(
    "T",
)
U = TypeVar("U")

# class Range(Generic[T], Annotated[T, Field()]):


# def range(tp: type[T], **kwargs):
#     # TP = TypeVar("TP", bound=T)

#     # class Range(Annotated[type[U], Field(**kwargs)], Generic[U]): ...

#     return Range[T]

AT = Annotated[T, Field()]

A = Annotated[str, Field(min_length=3)]

RangedT = Annotated[T, Field()]


# def ranged(tp: type):
#     return Annotated[tp, Field()]


# class MyAnnot(Annotated[T, Field()]): ...


# R = Ranged[int]


def range(tp: type[T]):
    return Annotated[T, Field()]


if __name__ == "__main__":
    # r = ranged(int)
    a: MyAnnot[int] = 0
    # a: RangedT[int] = 1



# https://docs.pydantic.dev/latest/concepts/models/#private-model-attributes
import random
from pydantic import PrivateAttr

from expedantic import ConfigBase


class Config(ConfigBase):
    _my_private_setting: bool = False
    _private_seed: int = PrivateAttr(
        default_factory=lambda: random.randint(0, 2**32 - 1)
    )
    public_setting: str = "This is PUBLIC"


if __name__ == "__main__":
    # You can attain the private setting with the parser.
    config = Config.parse_args()

    # The private setting will not be contained in the config instance.
    print(config.model_dump())
    """
    {'public_setting': 'This is PUBLIC'}
    """



import unittest

from expedantic import ConfigBase


class Config(ConfigBase):
    batch_size: int = 512
    this_config_is_very_important: bool = False


class TestUnderscore(unittest.TestCase):
    def test_underscore(self):
        config = Config.parse_args(
            replace_underscore_to_hyphen=True, args=["--batch-size", "256"]
        )
        self.assertEqual(config.batch_size, 256)

        config = Config.parse_args(
            replace_underscore_to_hyphen=True,
            args=["--this-config-is-very-important"],
        )
        self.assertTrue(config.this_config_is_very_important)



from pathlib import Path
from typing import Annotated, Literal
from expedantic import ConfigBase, Field


class Config(ConfigBase):
    algorithm: Literal["TRPO", "PPO", "SAC", "TD3"] = "TRPO"
    target_kl: Annotated[
        float,
        Field(
            title="Target Kullback-Leibler divergence between updates.",
            description="Should be small for stability. Values like 0.01, 0.05.",
            gt=0.0,
        ),
    ] = 0.01


config = Config()
Path("configs").mkdir(exist_ok=True)
config.save_as_yaml("configs/config.yaml")
Config.generate_schema("schemas/config_schema.json")



# import unittest
# from typing import Any, Literal
#
# from expedantic import ConfigBase, Field
#
#
# class TestInnerConfig(ConfigBase):
#     a: int = -1
#     b: float = -2.0
#     # req: str
#
#
# class Obj: ...
#
#
# class TestConfig(ConfigBase):
#     a: int = 1
#     b: float = 2.0
#     c: str = "3.0"
#     d: list[float] = Field(default_factory=lambda: [1.0, 2.0, 3.0])
#     e: TestInnerConfig = TestInnerConfig()
#     f: Any = "asdf"
#     g: Literal["A", "B", "C"]
#     h: list[int]
#     file: str
#     obj: Obj = Obj()
#
#
# class TestBase(ConfigBase):
#     a: int = 1
#     b: float = 2.0
#
#
# class TestChild(TestBase):
#     c: str = "c"
#
#
# # class TestSaveLoad(unittest.TestCase):
# # def test_save(self):
# # TestConfig.model_dump()
#
#
# if __name__ == "__main__":
#     # t = TestConfig()
#     # tapified = TestConfig.tapify()
#     # c = TestConfig.parse_args()
#     c = TestConfig(g="A", h=[1, 2, 3], file="test.png")
#     # TestConfig.generate_schema("test_schema.json")
#
#     from pathlib import Path
#
#     c.save_as_yaml("test.yaml")
#
#     print(c)



import argparse
import inspect
import json

from abc import ABC
from collections.abc import Mapping
from io import IOBase
from pathlib import Path
from types import UnionType
from typing import Any, Callable, Type, Literal, Union, get_origin, get_args
from typing_extensions import Self

import pydantic
import pydantic_yaml

# from ruamel.yaml import YAML
from ccorp.ruamel.yaml.include import YAML

pydantic_yaml._internals.v2.YAML = YAML

from . import utils


class NOT_PROVIDED_CLASS:
    def __repr__(self) -> str:
        return "NOT_PROVIDED"


_NOT_PROVIDED = NOT_PROVIDED_CLASS()


class ConfigBase(pydantic.BaseModel, Mapping, ABC):
    model_config = pydantic.ConfigDict(
        extra="forbid", protected_namespaces=("model_", "expedantic_")
    )

    _mutually_exclusive_sets: list[set[str]] = []
    """
    Defines sets of configuration options within `ConfigBase` that are mutually exclusive. Each set contains 
    keys of options where, logically, only one option can be evaluated as True at any given time. An option 
    is considered to be evaluated as True if its value is truthy (i.e., not `None`, `False`, an empty string, 
    or any other value that is considered falsy in a boolean context).

    If more than one option in a mutually exclusive set is evaluated as True, a `ValidationError` is raised, 
    enforcing that incompatible configuration states do not co-occur. This feature is crucial for maintaining 
    the integrity and logical consistency of the configuration.

    Type:
        list[set[str]]: Each inner set comprises strings that represent the keys of the configuration options. 
        These options are mutually exclusive with each other, ensuring that only one can be evaluated as True, 
        promoting clear and logical configuration setups.
    """

    def flatten(self, sep="."):
        """
        Flattens the configuration object's hierarchy into a single-level dictionary, using a
        custom separator for keys. This method is useful for converting nested configurations into
        a flat structure, facilitating serialization or simplifying access to nested values.

        Parameters:
        - sep (str, optional): The separator character used to join nested keys in the resulting
        flat dictionary. Defaults to '.'.

        Returns:
        dict[str, Any]: A flattened dictionary where keys are paths to the original nested values,
        joined by the specified separator. Each key corresponds to a value from the original
        configuration.

        Example:
        Given a configuration object structured as `{'outer': {'inner': 'value'}}` and using the
        default separator, the result would be `{'outer.inner': 'value'}`. If a separator of '/'
        is used, the result would be `{'outer/inner': 'value'}`.

        Note:
        The method does not modify the original configuration object but returns a new dictionary
        with the flattened structure.
        """
        return utils.flatten_dict(self.model_dump(), sep=sep)

    def compatible_args(self, cls: Type | Callable, *exclusive_keys: str):
        """
        Returns a dictionary of arguments from this config instance that match the signature of `cls`,
        excluding any specified in `exclusive_keys`. This method is useful for dynamically initialising
        components with applicable settings.

        Parameters:
        - cls (Type | Callable): Class or callable to filter compatible configuration arguments for.
        - exclusive_keys (str): Names of keys to exclude from the result.

        Returns:
        Dictionary of compatible arguments, excluding those listed in `exclusive_keys`.

        Example:
        Given a configuration with attributes 'a', 'b', 'c', and a function requiring 'a' and 'b',
        calling `self.compatible_args(target_function, 'c')` yields {'a': value, 'b': value}.
        """
        fields = self.model_dump()
        args = utils.get_kwargs(cls)
        arg_keys = set(args.keys())
        arg_keys -= set(exclusive_keys)
        return {k: v for k, v in fields.items() if k in arg_keys}

    def save_as_yaml(
        self,
        file: Path | str | IOBase,
        default_flow_style: bool | None = False,
        indent: int | None = None,
        map_indent: int | None = None,
        sequence_indent: int | None = None,
        sequence_dash_offset: int | None = None,
        custom_yaml_writer: YAML | None = None,
        **json_kwargs,
    ):
        """Write a YAML file representation of the model.

        Parameters
        ----------
        file : Path or str or IOBase
            The file path or stream to write to.
        model : BaseModel
            The model to write.
        default_flow_style : bool
            Whether to use "flow style" (more human-readable).
            https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
        indent : None or int
            General indent value. Leave as None for the default.
        map_indent, sequence_indent, sequence_dash_offset : None or int
            More specific indent values.
        custom_yaml_writer : None or YAML
            An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
            The above options will be set on it, if given.
        json_kwargs : Any
            Keyword arguments to pass `model.json()`.

        Notes
        -----
        This currently uses JSON dumping as an intermediary.
        This means that you can use `json_encoders` in your model.
        """
        pydantic_yaml.to_yaml_file(
            file,
            self,
            default_flow_style=default_flow_style,
            indent=indent,
            map_indent=map_indent,
            sequence_indent=sequence_indent,
            sequence_dash_offset=sequence_dash_offset,
            custom_yaml_writer=custom_yaml_writer,
            **json_kwargs,
        )

    @classmethod
    def generate_schema(cls, path: str | Path):
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        json.dump(cls.model_json_schema(), path.open("w"), indent=2)
        print(f"JSON Schema for {cls.__name__} is generated at {path}")

    @classmethod
    def load_from_yaml(cls, path: Path | str):
        path = Path(path)
        return pydantic_yaml.parse_yaml_file_as(cls, path)

    @classmethod
    def parse_args(
        cls,
        require_default_file: bool = False,
        replace_underscore_to_hyphen: bool = False,
        args=None,
        sep=".",
    ):
        """
        Parses command line arguments and returns an instance of the class with properties set
        based on the provided arguments. This method supports nested configurations and allows
        specifying complex data types including lists, sets, tuples, and literals. It also supports
        loading default values from a YAML file if `require_default_file` is set to True.

        Parameters:
        - require_default_file (bool, optional): If True, expects a path to a YAML file as the first
        positional argument from which to load default values. Defaults to False.
        - args (List[str], optional): A list of strings representing the arguments to parse. If None,
        parses arguments from sys.argv. Defaults to None.
        - sep (str, optional): The separator used for nested argument names. Defaults to '.'.

        Returns:
        An instance of the class, initialised with properties set based on the parsed arguments and
        any defaults specified in the YAML file (if `require_default_file` is True).

        This method dynamically constructs an argument parser based on the class's model fields,
        supporting deep nesting by using the specified separator `sep` to denote hierarchy in argument
        names. It handles various data types appropriately and allows for the specification of default
        values directly in the command line invocation or through a configuration file.
        """

        def _parse_params(
            parser: argparse.ArgumentParser,
            cls: Type[ConfigBase],
            parent_key: str = "",
            sep=sep,
        ):
            for name, field_info in cls.model_fields.items():
                if parent_key:
                    name = f"{parent_key}{sep}{name}"

                tp = field_info.annotation
                assert tp is not None
                origin = get_origin(tp)

                if not origin and inspect.isclass(tp) and issubclass(tp, ConfigBase):
                    _parse_params(parser, tp, name, sep)
                    continue

                if replace_underscore_to_hyphen:
                    name = name.replace("_", "-")

                names = [f"--{name}"]
                kwargs = {}
                kwargs["type"] = tp

                annot_repr = str(tp) if origin else tp.__name__

                if tp is Any:
                    kwargs["type"] = str
                elif tp is bool:
                    kwargs["action"] = argparse.BooleanOptionalAction
                elif origin is Literal:
                    var_type, literals = utils.get_literals(tp, name)
                    kwargs["type"] = var_type
                    kwargs["choices"] = literals
                    annot_repr = annot_repr.replace("typing.", "")
                elif origin in {list, set, tuple}:
                    kwargs["nargs"] = "*"
                    args = get_args(tp)
                    if len(args) == 0:
                        kwargs["type"] = str
                    else:
                        kwargs["type"] = args[0]
                elif origin in {UnionType, Union}:
                    kwargs["type"] = (
                        str  # Try to solve union types via pydantic internal.
                    )
                    if origin is Union:
                        args: tuple[type, ...] = get_args(tp)
                        annot_repr = " | ".join(map(lambda a: a.__name__, args))
                elif origin is dict:
                    kwargs["type"] = YAML().load

                if field_info.is_required():
                    req_repr = "required"
                    if not require_default_file:
                        kwargs["default"] = _NOT_PROVIDED
                else:
                    default_value = field_info.get_default(call_default_factory=True)
                    req_repr = f"default={default_value}"
                    if not require_default_file:
                        kwargs["default"] = default_value

                kwargs["help"] = f"({annot_repr}, {req_repr})"

                if require_default_file:
                    kwargs["default"] = _NOT_PROVIDED

                parser.add_argument(*names, **kwargs)

        parser = argparse.ArgumentParser()
        if require_default_file:
            parser.add_argument(
                "_config_file_path", type=Path, help="Default Config File Path"
            )
        _parse_params(parser, cls)
        args = parser.parse_args(args=args)

        if require_default_file:
            with open(args._config_file_path, "r") as f:
                file_dict = YAML().load(f)
            args.__dict__.pop("_config_file_path")
        else:
            file_dict = {}

        nested_args_dict = {}
        for k, v in vars(args).items():
            if v is _NOT_PROVIDED:
                continue
            keys = k.split(sep)
            current_level = nested_args_dict
            for key in keys[:-1]:
                current_level = current_level.setdefault(key, {})
            current_level[keys[-1]] = v

        def update_recursively(d: dict, other: dict):
            for k, v in other.items():
                if isinstance(v, dict):
                    d[k] = update_recursively(d.get(k, {}), v)
                else:
                    d[k] = v
            return d

        file_dict = update_recursively(file_dict, nested_args_dict)
        instance = cls.model_validate(file_dict)

        return instance

    @pydantic.model_validator(mode="after")
    def check_mutually_exclusive_sets(self) -> Self:
        for exclusive_set in self._mutually_exclusive_sets:
            check = sum(bool(self[key]) for key in exclusive_set) == 1
            if not check:
                raise ValueError(
                    f"Mutual exclusivity has broken. (set: {exclusive_set})"
                )
        return self

    def __getitem__(self, key: str):
        if key not in self.model_fields:
            raise KeyError(f"Key '{key}' not found.")
        return self.__getattribute__(key)

    def __len__(self):
        return len(self.model_fields)

    def keys(self):
        return self.model_fields.keys()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ccorp-yaml-include-relative-path",
    version="0.1.0",
    author="Yuri A. Mikhaylov",
    author_email="it@mikhailov.xyz",
    description="Exactly ccorp-yaml-include package with a bug fix and support of relative paths for included aliases",
    long_description_content_type="text/x-rst",
    long_description=long_description,
    url="https://github.com/mrkram/ccorp_yaml_include",
    packages=['ccorp.ruamel.yaml.include'],
    install_requires=['ruamel.yaml'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import unittest

from pydantic import ConfigDict, ValidationError
from pydantic_yaml import parse_yaml_raw_as
from ruamel.yaml import YAML

from expedantic import ConfigBase, Field


class Config(ConfigBase):
    model_config = ConfigDict(revalidate_instances="always")
    a: int = Field(1, gt=0)
    b: float = Field(0.5, lt=1.0)
    c: float = Field(0.05, le=1.0, ge=0.0)


class ConstraintsTest(unittest.TestCase):
    def test_int_float(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "a: 1.5")

    def test_gt(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "a: -1\nb: 0.9")

    def test_lt(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "a: 1\nb: 1.0")

    def test_le(self):
        config = parse_yaml_raw_as(Config, "c: 1.0")
        self.assertEqual(config.c, 1.0)

    def test_ge(self):
        config = parse_yaml_raw_as(Config, "c: 0.0")
        self.assertEqual(config.c, 0.0)
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "c: -0.1")


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path

from ccorp.ruamel.yaml.include import YAML
from expedantic import ConfigBase, Field


class Base(ConfigBase):
    device: str = "cpu"
    learning_rate: float = 3.0e-4
    batch_size: int = 1024


class Child(Base):
    gradient_clip_range: float = 0.2


class GrandChild(Child):
    decoder_net_arch: list[int] = [128, 128]


class TestInclude(unittest.TestCase):
    def setUp(self) -> None:
        base = """\
device: cpu
batch_size: 1024
learning_rate: 0.0003
"""
        child = """\
<<: !include ./tmp/base.yaml
device: 'cuda'
gradient_clip_range: 0.2
"""
        grandchild = """\
<<: !include ./tmp/child.yaml
learning_rate: 5.0e-5
gradient_clip_range: 0.3
decoder_net_arch:
- 128
- 128
"""
        yaml = YAML()

        self.base_file = Path("./tmp/base.yaml")
        self.child_file = Path("./tmp/child.yaml")
        self.grand_child_file = Path("./tmp/grand_child.yaml")

        self.base_file.parent.mkdir(exist_ok=True)
        with self.base_file.open("w") as f:
            yaml.dump(yaml.load(base), f)
        with self.child_file.open("w") as f:
            yaml.dump(yaml.load(child), f)
        with self.grand_child_file.open("w") as f:
            yaml.dump(yaml.load(grandchild), f)

    def tearDown(self) -> None:
        self.base_file.unlink()
        self.child_file.unlink()
        self.grand_child_file.unlink()

    def test_include(self):
        config = Base.load_from_yaml(self.base_file)
        self.assertEqual(config.device, "cpu")
        self.assertEqual(config.batch_size, 1024)
        self.assertEqual(config.learning_rate, 3.0e-4)

        config = Child.load_from_yaml(self.child_file)
        self.assertEqual(config.device, "cuda")
        self.assertEqual(config.gradient_clip_range, 0.2)

        config = GrandChild.load_from_yaml(self.grand_child_file)
        self.assertEqual(config.device, "cuda")
        self.assertEqual(config.learning_rate, 5.0e-5)
        self.assertEqual(config.gradient_clip_range, 0.3)
        self.assertEqual(config.decoder_net_arch, [128, 128])


if __name__ == "__main__":
    unittest.main()



import unittest

from pydantic import ValidationError

from expedantic import ConfigBase


class Config(ConfigBase):
    _mutually_exclusive_sets = [{"use_mlp_model", "use_batch_norm_on_cnns"}]

    use_mlp_model: bool = True
    use_batch_norm_on_cnns: bool = False


class TestMutuallyExclusiveSets(unittest.TestCase):
    def test_mutually_exclusive_bools(self):
        with self.assertRaises(ValidationError):
            config = Config(use_mlp_model=True, use_batch_norm_on_cnns=True)

        with self.assertRaises(ValidationError):
            config = Config(use_mlp_model=False, use_batch_norm_on_cnns=False)

        config = Config(use_mlp_model=False, use_batch_norm_on_cnns=True)
        self.assertEqual(config.use_mlp_model, False)
        self.assertEqual(config.use_batch_norm_on_cnns, True)



import unittest
import argparse
from typing import Any

from expedantic import ConfigBase


class Config(ConfigBase):
    class Inner(ConfigBase):
        name: str = "Inner"
        favourite_number: float | int = 3.14

    inner: Inner = Inner()
    optional: int | None = 1
    numbers: list[int] = [1, 2, 3]
    my_dict: dict[str, Any] = {}


class TestArgParse(unittest.TestCase):
    def test_arg_parse(self):
        config = Config.parse_args(args=["--optional", "3"])
        self.assertEqual(config.optional, 3)

        config = Config.parse_args(args=["--numbers", "5", "6", "7"])
        self.assertEqual(config.numbers, [5, 6, 7])

        config = Config.parse_args(
            args=["--inner.name", "Test", "--inner.favourite_number", "7"]
        )
        self.assertEqual(config.inner.name, "Test")
        self.assertEqual(config.inner.favourite_number, 7)

    def test_parse_dict(self):
        config = Config.parse_args(args=["--my_dict", "{key: value}"])
        self.assertEqual(config.my_dict.get("key"), "value")



import unittest
from typing import Literal

from pydantic import ValidationError
from pydantic_yaml import parse_yaml_raw_as

from expedantic import ConfigBase


class Config(ConfigBase):
    distribution: Literal["Gaussian", "Uniform", "Cauchy", "Poisson"]


class TestLiteral(unittest.TestCase):
    def test_literal(self):
        with self.assertRaises(ValidationError):
            config = parse_yaml_raw_as(Config, "distribution: Dirichlet")

        config = parse_yaml_raw_as(Config, "distribution: Gaussian")
        self.assertEqual(config.distribution, "Gaussian")



import importlib
import inspect
from dataclasses import field, dataclass
from typing import Annotated, Any, Generic, TypeVar, Callable, ClassVar, get_args
from typing_extensions import Self
from types import ModuleType, new_class

# from pydantic.fields import

from pydantic_core import CoreSchema, core_schema
from pydantic import (
    BaseModel,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    TypeAdapter,
    Field,
    computed_field,
    model_validator,
    ValidationError,
    GetPydanticSchema,
)

# from pydantic.dataclasses import dataclass
from pydantic import InstanceOf
from pydantic.functional_validators import (
    BeforeValidator,
    PlainValidator,
    AfterValidator,
    WrapValidator,
)

from expedantic import ConfigBase

T = TypeVar("T")


# @dataclass
# class ModuleObjectModel(Generic[T]):
#     _type: type[T]
#     name: str
#     module: str
#     positional_arguments: list = field(default_factory=lambda: [])
#     keyword_arguments: dict[str, Any] = field(default_factory=lambda: {})

#     def get_instance(self) -> T:
#         module = importlib.import_module(self.module)
#         attr = getattr(module, self.name)

#         assert issubclass(type(attr), self._type)

#         if inspect.isclass(attr):
#             return attr(*self.positional_arguments, **self.keyword_arguments)
#         else:
#             return attr


# from expedantic import ConfigBase


# T = TypeVar("T", bound=ModuleType)


class ModuleObjectModel(BaseModel):
    # _type: type[T] = None
    name: str
    module: str
    positional_arguments: list = field(default_factory=lambda: [])
    keyword_arguments: dict[str, Any] = field(default_factory=lambda: {})
    # _type: type[T]

    # @computed_field
    # @classmethod
    # def _type(cls) -> type[T]:
    #     return cls.__orig_bases__[1]

    # @model_validator(mode="before")
    # @classmethod
    # def get_type(cls, values) -> type[T]:
    #     cls._type = get_args(cls.__orig_bases__[1])
    #     return values

    # def get_instance(self) -> T:
    #     module = importlib.import_module(self.module)
    #     attr = getattr(module, self.name)

    #     # assert issubclass(type(attr), _typ)

    #     if inspect.isclass(attr):
    #         return attr(*self.positional_arguments, **self.keyword_arguments)
    #     else:
    #         return attr

    # @staticmethod
    # def get_validator(_type: type):
    #     def get_instance(self) -> T:
    #         module = importlib.import_module(self.module)
    #         attr = getattr(module, self.name)

    #         assert issubclass(
    #             type(attr), _type
    #         ), f"{self.module}.{self.name} is not a type of {_type}"

    #         if inspect.isclass(attr):
    #             return attr(*self.positional_arguments, **self.keyword_arguments)
    #         else:
    #             return attr

    #     return get_instance

    @model_validator(mode="after")
    def get_instance(self):
        module = importlib.import_module(self.module)
        attr = getattr(module, self.name)

        # assert issubclass(type(attr), _typ)

        if inspect.isclass(attr):
            return attr(*self.positional_arguments, **self.keyword_arguments)
        else:
            return attr


# def ModuleObject(_type: type[T]):
#     return Annotated[
#         T | ModuleObjectModel, PlainValidator(ModuleObjectModel.get_validator(_type))
#     ]


ModuleObject = Annotated[
    T | ModuleObjectModel,
    Annotated[ModuleObjectModel, ...],
    BeforeValidator(ModuleObjectModel.get_instance),
    InstanceOf[T],
]


@dataclass
class ModuleObjectDict:
    name: str
    module: str
    positional_arguments: list = field(default_factory=lambda: [])
    keyword_arguments: dict[str, Any] = field(default_factory=lambda: {})

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ):
        def get_instance(d):
            module = importlib.import_module(d["module"])
            attr = getattr(module, d["name"])

            # assert issubclass(type(attr), _typ)

            if inspect.isclass(attr):
                return attr(*d["positional_arguments"], **d["keyword_arguments"])
            else:
                return attr

        from_dict_schema = core_schema.chain_schema(
            [
                core_schema.dataclass_schema(),
                core_schema.typed_dict_schema(
                    {
                        "name": core_schema.typed_dict_field(core_schema.str_schema()),
                        "module": core_schema.typed_dict_field(
                            core_schema.str_schema()
                        ),
                        "positional_arguments": core_schema.typed_dict_field(
                            core_schema.list_schema()
                        ),
                        "keyword_arguments": core_schema.typed_dict_field(
                            core_schema.dict_schema(
                                core_schema.str_schema(), core_schema.any_schema()
                            )
                        ),
                    }
                ),
                core_schema.no_info_plain_validator_function(
                    get_instance
                    # ModuleObjectModel.get_instance
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=core_schema.union_schema([from_dict_schema]),
        )


# class _ModuleObjectModelValidator:
#     @classmethod
#     def __get_pydantic_core_schema__(
#         cls, source_type: Any, handler: GetCoreSchemaHandler
#     ) -> CoreSchema:

#         from_dict_schema = core_schema.chain_schema(
#             [
#                 core_schema.typed_dict_schema(
#                     {
#                         "name": core_schema.typed_dict_field(core_schema.str_schema()),
#                         "module": core_schema.typed_dict_field(
#                             core_schema.str_schema()
#                         ),
#                         "positional_arguments": core_schema.typed_dict_field(
#                             core_schema.list_schema()
#                         ),
#                         "keyword_arguments": core_schema.typed_dict_field(
#                             core_schema.dict_schema(
#                                 core_schema.str_schema(), core_schema.any_schema()
#                             )
#                         ),
#                     }
#                 ),
#                 core_schema.no_info_plain_validator_function(
#                     ModuleObjectModel[T].get_instance
#                 ),
#             ]
#         )

#         return core_schema.json_or_python_schema(
#             json_schema=from_dict_schema,
#             python_schema=core_schema.union_schema(
#                 [core_schema.is_instance_schema(ModuleObjectModel[T]), from_dict_schema]
#             ),
#         )

#         return core_schema.no_info_before_validator_function(
#             self.get_instance, handler.generate_schema(source_type)
#         )

#     @classmethod
#     def __get_pydantic_json_schema__(
#         cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
#     ):
#         return handler(core_schema.any_schema())


# class _ModuleObjectAlias(type):
#     def __new__()
#     ...


# class ModuleObject(Any):
#     @staticmethod
#     def validate(cls, v):
#         if not isinstance(v, str):
#             raise TypeError("String required.")

#         if hasattr(cls._module, v):
#             attr = getattr(cls._module, v)
#             if cls._type is Any or isinstance(attr, cls._type):
#                 return attr
#             else:
#                 raise ValueError(f"{v} is not a type of {cls._type}.")
#         else:
#             raise ValueError(f"{v} is not a part of module {self._module}.")

#     def __class_getitem__(cls, key: tuple[ModuleType, Any] | ModuleType) -> type[str]:
#         if isinstance(key, tuple):
#             key, typ = key
#         else:
#             key, typ = key, Any

#         base_type = tuple() if typ is Any else (typ,)

#         def exec_body(ns: dict[str, Any]):
#             ns.__setitem__("_module", key)
#             ns.__setitem__("_type", typ)
#             ns.__setitem__("validate", classmethod(ModuleObject.validate))
#             # ns.__setitem__("_type", typ)

#         return new_class(
#             f"{cls.__name__}[{key.__name__}]",
#             (cls, str),
#             {},
#             exec_body,
#         )
#         # class ModuleObject(type):
#         #     _module: ModuleType = key

#         #     # def __new__(cls):
#         #     #     return new_class(
#         #     #         f"{cls.__name__}[{key.__name__}]",
#         #     #         (cls,),
#         #     #         {},
#         #     #         lambda ns: ns.__setitem__("_module", key),
#         #     #     )

#         # return ModuleObject


# class TestModuleObject(metaclass=ModuleObjectMeta): ...


# class ModuleObject(str, Generic[T]):
#     def __init__(self, module: ModuleType) -> None:
#         super().__init__()
#         self._module = module

#     @classmethod
#     def __get_validators__(cls):
#         yield cls.validate

#     @classmethod
#     def validate(cls, v, field):
#         if not isinstance(v, str):
#             raise TypeError("string name required.")

#         if not


# class Config(ConfigBase): ...


if __name__ == "__main__":
    import math

    # # M = TestModuleObject[math]
    # M = ModuleObject[math, Callable]
    # print(M)
    # # print(M._module)
    # print(vars(M))

    # # m = ModuleObject["math"]

    # print(M.validate("sqrt"))

    # try:
    #     # math.pi
    #     FloatFromMath = ModuleObject[math, float]
    #     pi = FloatFromMath.validate("pi")
    #     assert isinstance(pi, float)
    #     sqrt = FloatFromMath.validate("sqrt")
    # except Exception as e:
    #     print(e)

    # class TestStr(str): ...

    # class Config(ConfigBase):
    #     a: Annotated[str, ModuleObject[math, float]] = "pi"
    #     b: ModuleObject[math, float] = "pi"

    # c = Config()
    # print(c)

    # FloatFromMath = Annotated[str, ModuleObjectModel("pi", "math", float)]
    # FloatFromMath = Annotated[float, _ModuleObjectModelValidator[float]]
    # FloatFromMath = ModuleObject[T]

    # class Config(ConfigBase):
    #     a: FloatFromMath
    #     b: ModuleObject[int]

    # try:
    #     c = Config(
    #         a=ModuleObjectModel(name="pi", module="math"),
    #         b=ModuleObjectModel(name="pi", module="math"),
    #     )
    # except ValidationError as e:
    #     print(e)

    # FloatFromMath = Annotated[float, ModuleObjectDict("pi", "math")]

    import pydantic_yaml

    # def model_object():
    #     return GetPydanticSchema(
    #         lambda tp, handler: core_schema.union_schema(
    #             core_schema.is_instance_schema()
    #         )
    #     )

    MOM = Annotated[T | ModuleObjectModel, ...]

    # def check

    # MOM = Annotated[T, WrapValidator()]

    class Config2(ConfigBase):
        # a: ModuleObject[ConfigBase]
        # a: FloatFromMath
        a: ModuleObjectModel | float
        b: float | ModuleObjectModel = ModuleObjectModel(name="pi", module="math")
        c: Annotated[
            float | ModuleObjectModel,
            Field(ModuleObjectModel(name="pi", module="math"), validate_default=True),
        ]
        d: float | ModuleObjectModel = 3.14
        e: float = ModuleObjectModel(name="pi", module="math")
        # f: Annotated[
        #     float,
        #     Field(ModuleObjectModel(name="pi", module="math"), validate_default=True),
        # ]
        g: MOM[float]
        h: Annotated[MOM[float], Field(2.78)]

    c = Config2(
        a=ModuleObjectModel(name="pi", module="math"),
        b=ModuleObjectModel(name="pi", module="math"),
        g=ModuleObjectModel(name="e", module="math"),
    )
    print(c)
    print(c.model_dump_json())

    # import numpy

    # ModuleObjectModel.model_validate_json('{"name": }')

    # c = Config2(a={"name": "pi", "module": "math"})
    # print(c)

    # class Config3(ConfigBase):
    #     # a: Annotated[float | ModuleObjectModel] = ModuleObjectModel(
    #     # name="pi", module="math"
    #     # )
    #     a: Annotated[float] = ModuleObjectModel(
    #         name="pi", module="math"
    #     )

    # c = Config3()
    # print(c)

    # c = pydantic_yaml.parse_yaml_raw_as(
    #     Config2, "a: {name: ConfigBase, module: expedantic}"
    # )



# https://docs.pydantic.dev/latest/concepts/unions/#discriminated-unions
from pathlib import Path
from typing import Literal

from expedantic import ConfigBase, Field


class GANConfig(ConfigBase):
    name: Literal["GAN"] = "GAN"
    learning_rate_generator: float = 1.0e-4
    learning_rate_discriminator: float = 1.0e-5
    batch_size: int = 128
    n_epochs: int = 10
    latent_size: int = 16


class WGANConfig(GANConfig):
    name: Literal["WGAN"] = "WGAN"
    weight_clipping_range: float = 0.01
    n_critic_epochs: int = 5


class WGANGPConfig(WGANConfig):
    name: Literal["WGAN_GP"] = "WGAN_GP"
    gp_lambda: float = 10


class Config(ConfigBase):
    device: Literal["cpu", "cuda"] = "cpu"
    dataset: Literal["mnist", "cifar10"] = "mnist"
    checkpoint_freq: int = 1000
    checkpoint_save_dir: Path = Path("./checkpoints/")
    verbose: bool = True
    gan_config: GANConfig | WGANConfig | WGANGPConfig = Field(..., discriminator="name")


if __name__ == "__main__":
    config = Config(gan_config={"name": "WGAN"})  # type: ignore
    print(config.gan_config)
    """
    name='WGAN' learning_rate_generator=0.0001 
    learning_rate_discriminator=1e-05 batch_size=128
    n_epochs=10 latent_size=16 weight_clipping_range=0.01
    n_critic_epochs=5
    """
